#!/usr/bin/perl

# Note that in this program, I invented the term "cookie bag".  I consider the cookie "jar" for a 
# given user to be all of the cookies in all of his .cookies files.  Each different domain restriction 
# (e.g. .google.com) has it's own file and I'm calling that file a cookie "bag" since it is a smaller 
# collection of cookies than the whole "jar".

# Also note that I have made no attempt to handle cookies in compliance with any RFC as they have 
# all taken a good, simple thing and made way too much of it.  I instead have attempted to comply 
# (as much as is practicable) with the Netscape Cookie Specification found here 
# http://wp.netscape.com/newsref/std/cookie_spec.html.

# Particular thanks to Philippe "BooK" Bruhat for the Perl HTTP::Proxy module which is the core of 
# HoTTProxy.

# To-do List:
# ===========
# Need to finish the web based administrative control panel
# Discard expired cookies (get rid of them in the cookie bags) instead of just not sending them
# Need to do a filefree check before opening cookie bag file for output so as not to step on self
# Need to do something with the logging so that the log file wraps off or something
# Need to make CookiePassthru and SetCookiePassthru do something
# Need to write the manual

# Note: HoTTProxy started out as a quick hack to see if I could make a proxy server. Once it worked 
#       (suprisingly easily mind you), then I set out to add the proxy cookie support, and more and 
#       more features. Since it grew kind of quickly, and without any thought of others seeing the 
#       code, some things in here might still be a bit kludgy. One notable example would be that 
#       there are still a bunch of remarked out "print STDERR something" lines. I use these for 
#       troubleshooting and un-rem them as necessary. Some of these have been turned into "say" 
#       instructions and given a ConsoleVerbosity level, but many have not. The code *is* fairly 
#       well commented, so if you are skilled with Perl you should be able to follow it.

# We need 0.16 or greater because the ability to add protocols to protocols_allowed wasn't added until 0.16
use HTTP::Proxy 0.16 qw( :log );

use HTTP::Proxy::HeaderFilter::simple;
use HTTP::Proxy::BodyFilter::simple;
use HTTP::Date;	# We use this module when converting epoch time to "cookie" time and vice versa
use MIME::Base64 qw( encode_base64 decode_base64 );
use IO::Handle;
use LWP::UserAgent;
use Digest::MD5 qw / md5_hex /;
use Date::Calc qw / Date_to_Time Decode_Date_US /;
use Socket;		# To figure out the local host name
use Sys::Hostname;		# To figure out the local host name
use File::Glob qw / :glob /;

require 'HoTTProxy_Add-In_Filters.pl';		# Any user-defined filters can be put in here

my $HoTTProxyVersion = '0.24.0.0';

use strict;

$|=1;

# We need the path to the present executable.  Under Win32 this is pretty 
# positive using the Win32::GetFullPathName.  Under Linux and the like, 
# it looks like $ENV{'PWD'} will get us what we want - suggestions?
my $path;
my $slash;
if ($^O eq 'MSWin32') {
	$path = Win32::GetFullPathName( $0 );
	($path) = $path =~ m/^(.*)\\/;	# e.g. C:\data\programming\Perl\HoTTProxy (no trailing slash)
	
} else {
	$path = $ENV{'PWD'};
}

my %HoTTProxyConfig;		# This hash will store the configuration settings for the proxy
my %hiddenConfig;			# Items listed in this hash won't show on the console - start the element name with a minus ('-') to hide - mostly used for items you just don't care about

say("
HoTTProxy $HoTTProxyVersion - (c) 2005 Brian A. Blakley

This program is free for personal use but must be licensed for commercial use.

License is contained in License.txt. License.txt is only a valid file if it is
signed by PGP Key \"Brian Blakley <bblakley\@mp5.net> (0xE66620C3)\".

",0);

processParameters();		# Handle all the command line parameters

# Set default config file if one not specified on command line
$HoTTProxyConfig{'ConfigFile'} = 'HoTTProxy.conf' unless $HoTTProxyConfig{'ConfigFile'};

loadConfig($HoTTProxyConfig{'ConfigFile'});		# Load up all the customizable settings

# Yes, I know we already did this, but some command line parameters *over-ride* the 
# config file while some need to be handled before we load the config file.  This is a 
# cheap and easy way to make this happen.
processParameters();		

$HoTTProxyConfig{'RunningFrom'} = $path;	# This is just to get the path to show on the console - we use $path most places in the code

# Yes, I know Perl doesn't care, but since this item shows on the console, it confuses some Win32 users 
# to see a frontslash in the middle of their path.  We don't want confused users do we?
if ($^O eq 'MSWin32') {
	$HoTTProxyConfig{'CookieDirectory'} = "$path\\$HoTTProxyConfig{'CookieDirectory'}";
} else {
	$HoTTProxyConfig{'CookieDirectory'} = "$path/$HoTTProxyConfig{'CookieDirectory'}";
}

# Find the "public" IP Address of this machine
if ($HoTTProxyConfig{'IPDetectionURL'} and $HoTTProxyConfig{'IPDetectionRegEx'}) {
	my $ua = LWP::UserAgent->new(agent=>"HoTTProxy/$HoTTProxyVersion");
	my $response = $ua->get($HoTTProxyConfig{'IPDetectionURL'});
	my $content = $response->content;
	if ($content) {
		eval("(\$HoTTProxyConfig{'PublicIPAddr'}) = \$content =~ $HoTTProxyConfig{'IPDetectionRegEx'}");
	}
}

# Find the "private" IP Address of this machine
$HoTTProxyConfig{'LocalHostName'} = hostname();
$HoTTProxyConfig{'LocalHostName'} = 'localhost' unless $HoTTProxyConfig{'LocalHostName'};
$HoTTProxyConfig{'PrivateIPAddr'} = inet_ntoa(scalar gethostbyname($HoTTProxyConfig{'LocalHostName'}));

# Determine whether or not Proxy Cookie Support is (or can be) enabled
if ( (-e "$HoTTProxyConfig{'CookieDirectory'}/") and ( eval("proxyCookieSupport()") ) and (not $HoTTProxyConfig{'DisableProxyCookieSupport'})) {
	$HoTTProxyConfig{'ProxyCookiesEnabled'} = 1;
} else {
	$HoTTProxyConfig{'ProxyCookiesEnabled'} = 0;
}

# I want to get rid of this $logDetail variable at some point and just use $HoTTProxyConfig{'LogDetail'}
my $logDetail = 1 if $HoTTProxyConfig{'LogDetail'};

say ("\nRun this program with the '/license' switch to see the license.\n",0);

say ("\n" . '=' x 79 . "\n\n",0);

foreach my $i (sort(keys(%HoTTProxyConfig))) {
	if (not $hiddenConfig{$i}) {
		if(ref($HoTTProxyConfig{$i}) eq 'HASH') {
			my @configKeys = sort(keys(%{$HoTTProxyConfig{$i}}));
			say(sprintf("%27s: %s","+$i",$configKeys[0]) . "\n",1);
			my $j;
			for ($j=1; $j<=$#configKeys; $j++) {
				say(sprintf("%27s  %s",'',$configKeys[$j]) . "\n",1);
			}
		} else {
			say(sprintf("%27s: %s",$i,$HoTTProxyConfig{$i}) . "\n",1);
		}
	}
}

say ("\n" . '=' x 79 . "\n\n",1);

my @userFiles = bsd_glob("$path/*.user");	# We just need a count of files - we don't do anything else with these two arrays
my @passwordFiles = bsd_glob("$path/*.password");

say("!!!!! Warning !!!!!\n\nNo .user or .password files found. You will be unable to access\nthis proxy until you create at least one user account.\n\nSee the documentation for details!\n\n") unless (scalar(@userFiles) or scalar(@passwordFiles));

my %cookieStore;		# This hash will hold all of the cookie bags

# I have no idea if this is correct or not, but I wanted to overload the HTTP::Daemon->proxy_tokens method so the 
# "Server" header that is returned is "HoTTProxy/#.#.#.#" instead of the default "libwww-perl-daemon/#.##"
# This seems to work - hopefully someone will tell me if this is bad practice.
{
	package HTTP::Daemon;
	sub product_tokens {
		return "HoTTProxy/$HoTTProxyVersion";
	}
}

my $proxy = HTTP::Proxy->new( port => $HoTTProxyConfig{'ProxyPort'}, host => '', via => $HoTTProxyConfig{'ViaString'} );	# Via is what shows up has HTTP_VIA on the web server

# ------------- This block of code is to enable capturing and redirecting of 'proxy:homepage' homepage requests as seen in Alltel
$proxy->init();
$proxy->agent->protocols_allowed( [ @{ $proxy->agent->protocols_allowed() }, 'proxy' ] );		# Add bogus protocol 'proxy' to the list of supported protocols
# ------------- End special code for proxy:homepage


if ($HoTTProxyConfig{'HTTP_Proxy_Module_Logmask'} = uc($HoTTProxyConfig{'HTTP_Proxy_Module_Logmask'})) {
	$HoTTProxyConfig{'HTTP_Proxy_Module_Logmask'} =~ s/[^A-Z |\w]//g;	# Only allow capital A-Z, spaces, and |
	eval "\$proxy->logmask($HoTTProxyConfig{'HTTP_Proxy_Module_Logmask'})";
}

if ($HoTTProxyConfig{'HTTP_Proxy_Module_Logfile'}) {
	open (HPML,">>$path/$HoTTProxyConfig{'HTTP_Proxy_Module_Logfile'}") or die "Couldn't open logfile $path/$HoTTProxyConfig{'HTTP_Proxy_Module_Logfile'}. Ending.\n   Error: $!\n";
	print HPML "\n" . sortableDate() . " HoTTProxy Ver. $HoTTProxyVersion Started up...\n\n";
	$proxy->logfh(*HPML);
}

open (OUT,">>$path/HoTTProxyLog.txt") or die "Couldn't open logfile $path/HoTTProxyLog.txt. Ending.\n   Error: $!\n";

# Record the startup in the log
logevent('SYSTEM','127.0.0.1',"HoTTProxy Ver. $HoTTProxyVersion Started up...");

# The request header filter that follows handles the authentication, and the 
# insertion of previously stored proxy cookies into the client supplied request 
# headers (if proxy cookies are enabled).  It also strips off any cookie header 
# that the browser tries to provide unless the CookiePassthru setting has said 
# to allow them through.
    
my $requestHeaderFilter = HTTP::Proxy::HeaderFilter::simple->new(
   sub {
      my ( $self, $headers, $request ) = @_;
      my $auth = $self->proxy->hop_headers->header('Proxy-Authorization') || "";
		my ($authUser, $user, $pw, $correctPw, $homePage, $expires, $expiresEpoch, $i, $j, $k, $hop, $cookieBagList, @cookies2send);
		
		@cookies2send = ();

		my $uri = scalar($$request{'_uri'});
		my ($protocol, $hoststring, $url, $params) = $uri =~ m/([^:]*):\/\/([^\/]*)(\/[^?]*)?(\?.*)?/;

		my $host = $$request{'_headers'}{'host'};
		
		# print STDERR "Raw host in request is: $$request{'_headers'}{'host'}\n";

	   # // ***************  Rule validation for host names  *************** \\
	   
	   # Host name rules:
	   
	   # 1. Name must start with a period
	   # 2. If name less than 2 periods then .local is appended
	
	   # Rule #1
	   if (not $host =~ m/^\./) {
	   	$host = '.' . $host;	# make it so
	   }
	   
	   # Rule #2
	   if (periodCount($host) < 2) {
	   	$host .= '.local';		# See RFC2109 to get an idea of what I'm doing here
	   }
	
	   # \\ ***************  Rule validation for host names  *************** //

      ($authUser) = $auth =~ m/^Basic (.*)$/;
		$authUser = decode_base64($authUser);
		# ($user,$pw) = split(/:/,$authUser);
		($user,$pw) = $authUser =~ m/^([^:]+):(.+)$/;

		if ($user =~ m/[ \\\/:+*?"<>|&`\x00-\x21]/) {	# Some characters we don't want in usernames
			$user = '';		
		}
		
		$correctPw = '';	# If $correctPw is still blank after authentication will fail
		if ($user) {
			my $filehandle;
			if (open($filehandle,"$path/$user.user")) {		# Note this file should contain the user's password, followed by CRLF followed by their choice of homepage URL (fully qualified http://host.domain.tld)
				$correctPw = <$filehandle>;
				chomp $correctPw;
				$homePage = <$filehandle>;
				chomp $homePage;
				$homePage = $HoTTProxyConfig{'DefaultHomePage'} unless $homePage;		# use the systemwide default homepage if the user doesn't have one
				$expires = <$filehandle>;
				chomp $expires;
				if (my ($year,$month,$day) = Decode_Date_US($expires)) {
					$expiresEpoch = Date_to_Time($year,$month,$day,0,0,0);	# Expiration is at midnight (the start of) of the expiration date
				}
				close $filehandle;
			} elsif (open($filehandle,"$path/$user.password")) {		# Note this file should contain the user's password, followed by CRLF followed by their choice of homepage URL (fully qualified http://host.domain.tld)
				$correctPw = <$filehandle>;
				chomp $correctPw;
				$correctPw = md5_hex($correctPw);	# Make it the same as if it had been an MD5 hash in the file to make comparing simple later
				$homePage = <$filehandle>;
				chomp $homePage;
				$homePage = $HoTTProxyConfig{'DefaultHomePage'} unless $homePage;		# use the systemwide default homepage if the user doesn't have one
				$expires = <$filehandle>;
				chomp $expires;
				if (my ($year,$month,$day) = Decode_Date_US($expires)) {
					$expiresEpoch = Date_to_Time($year,$month,$day,0,0,0);	# Expiration is at midnight (the start of) of the expiration date
				}
				close $filehandle;
			} else {
				$correctPw = '';		# Set $correctPw back to blank if we can't open the file - so auth will fail
				print STDERR sortableDate() . " Can't open password file $user.user or $user.password in $path\n   Error: $!\n";
			}
		}

		# check the credentials

		say ("User: $user, ExpiresEpoch is: $expiresEpoch and time is " . time . "\n",5);
		
		if (($expiresEpoch) and ($expiresEpoch < time)) {
			# account is expired - send 401 "unauthorized"
			if ($user) {
				logevent ($user,$$headers{'x-forwarded-for'},"Password is expired as of $expires");
				print STDERR sortableDate() . " $user Password is expired as of $expires\n";
			}
			if ($HoTTProxyConfig{'StealthExpiredAccount'}) {
	         my $response = HTTP::Response->new(407);
	         $response->header(
	            Proxy_Authenticate => "Basic realm=\"$HoTTProxyConfig{'RealmString'}\"" );
	            $self->proxy->response($response);
			} else {
				my $response = HTTP::Response->new(401);
         	$self->proxy->response($response);
         }
		} elsif (( md5_hex($pw) ne $correctPw ) or ( not $correctPw )) {
			# bad credentials - Don't send the real response - send a header requesting the browser to authenticate itself
			if ($user) {
				logevent ($user,$$headers{'x-forwarded-for'},"Failed auth trying password $pw");
				say("User $user failed auth trying password $pw\n",1);
			}
			
			if (($HoTTProxyConfig{'StealthInvalidSignon'}) or (not $user)) {	# Can't do this if no $user because the browser never auth's on the first try and we need to tell them to
	         my $response = HTTP::Response->new(407);
	         $response->header(
	            Proxy_Authenticate => "Basic realm=\"$HoTTProxyConfig{'RealmString'}\"" );
	            $self->proxy->response($response);
         } else {
				my $response = HTTP::Response->new(401);
         	$self->proxy->response($response);
	      }
      } elsif (($host =~ m/\.local$/) and (not $HoTTProxyConfig{'ValidLocalHosts'}{$$request{'_headers'}{'host'}}) and not ($HoTTProxyConfig{'DisableLocalHostRedirection'})) {		# if homepage is a .local domain
      	# credentials are good but we need to redirect
			say("User, IP: $user, $$headers{'x-forwarded-for'}\n",2);
      	say("Redirect: Redirecting $host to $homePage\n\n",2);
			my $response = HTTP::Response->new(307);
        	$response->header( Location => $homePage );
         $self->proxy->response($response);
      } else {
      	# good credentials
      	
			say(" User, IP: $user, $$headers{'x-forwarded-for'}\n",2);
			say("Host, URL: $host, $url",2);	# The \n's for this come later

			logevent($user,$$headers{'x-forwarded-for'},"Host: $$self{'_hphf_proxy'}{'request'}{'_headers'}{'host'}\t$$request{'_method'}\t" . scalar($$request{'_uri'}));

			# These shouldn't exist unless someone is trying something crafty, but we'll delete them to be safe
			delete $$headers{'HoTTProxyUser'} if defined $$headers{'HoTTProxyUser'};
			delete $$headers{'HoTTProxyToken'} if defined $$headers{'HoTTProxyToken'};

			if ($host =~ m/.local$/) {		# it is a local domain
				$$headers{'HoTTProxyUser'} = $user;
				$$headers{'HoTTProxyToken'} = $HoTTProxyConfig{'HoTTProxyToken'};
				say (" (Local domain)",3);
			} else {
				foreach my $i (keys(%{$HoTTProxyConfig{'TrustedDomains'}})) {
					if ($host =~ m/$i$/) {		# If host tail matches a trusted domain, it is trusted
						$$headers{'HoTTProxyUser'} = $user;
						$$headers{'HoTTProxyToken'} = $HoTTProxyConfig{'HoTTProxyToken'};
						say (" (Trusted domain)",3);
						last;
					}
				}
			}
			
			if (not defined $$headers{'HoTTProxyUser'}) {
				say(" (Non-local, non-trusted domain)",3);
			}
			
			say ("\n\n",2);

			# We're going to store the authenticated user name somewhere in the proxy object 
			# that is easy to get at so we can retrieve it on the response side so we know 
			# who the cookies are for.
			$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'} = $user;

			# Cookie processing goes here
			
			if ($HoTTProxyConfig{'ProxyCookiesEnabled'}) {
				# print STDERR "\nLooking here for cookie bags: $HoTTProxyConfig{'CookieDirectory'}/$user@*.cookies\n\n";
				foreach $i (bsd_glob("$HoTTProxyConfig{'CookieDirectory'}/$user@*.cookies")) {
					my ($domain) = $i =~ m/@(.*).cookies$/;
	
					# print STDERR "Cookie bag for domian $domain found ";
					if ($host =~ m/$domain$/) {		# If the domain of the cookie bag tail matches the host, it is applicable
						# print STDERR "which *does* need to be sent.\n";
						my $filehandle;
						if (open ($filehandle,$i)) {
							while (<$filehandle>) {
								chomp $_;
								# print STDERR "   Found: $_\n";
								# print STDERR "Time: $curtime\n";
								my %cookiehash = parseStoredCookie($_);
								# print STDERR "Path restriction: $cookiehash{'path'}\n";
								if ($url =~ m/^$cookiehash{'path'}/i) {		# if the cookie's path head matches the path of the script it should be sent
									# print STDERR "      which should be sent.\n\n";
									if ((not $cookiehash{'expiresEpoch'}) or ($cookiehash{'expiresEpoch'} >= time)) {		# Don't use expired cookies
										$cookies2send[$#cookies2send+1] = {%cookiehash};	# We're going to store the cookies we want to send here so we can sort them before we send them.
									}
								} else {
									# print STDERR "      which should not be sent.\n\n";
								}
							}
							close $filehandle;
							# print STDERR "\n"; 
						} else {
							print STDERR sortableDate() . " Tried to open cookie bag $i and failed.\n   Error: $!\n\n";
						}
					} else {
						# print STDERR "which does not need to be sent.\n\n";
					}
				}
	
				# foreach $i (@cookies2send) {
					# print STDERR "         $i: $$i{'name'} = $$i{'value'} ($$i{'domain'} $$i{'path'}) ($$i{'expires'} $$i{'expiresEpoch'})\n";
				# }
	
				# print STDERR "\n\n";
	
				@cookies2send = sort {length($$b{'path'}) <=> length($$a{'path'}) || length($$b{'domain'}) <=> length($$a{'domain'})} @cookies2send;	# cookies with more specific paths, or more specific domains if the paths are the same should go first - longer names are more specific names
	
				# foreach $i (@cookies2send) {
					# print STDERR "         $i: $$i{'name'} = $$i{'value'} ($$i{'domain'} $$i{'path'}) ($$i{'expires'} $$i{'expiresEpoch'})\n";
				# }
				
				my $cookieString = '';
				if (defined $cookies2send[0]) {
					$cookieString = join('=',($cookies2send[0]{'name'},$cookies2send[0]{'value'}));
				}
				for ($i=1; $i <= $#cookies2send; $i++) {
					$cookieString .= '; ' . join('=',($cookies2send[$i]{'name'},$cookies2send[$i]{'value'}));
				}
	
				if ($cookieString) {
					$$headers{'cookie'} = $cookieString;
					logevent ($user,$$headers{'x-forwarded-for'},"Host: $$self{'_hphf_proxy'}{'request'}{'_headers'}{'host'}\tCookie\t$cookieString") if $logDetail;
					say("Give cookie: $cookieString\n",9);
				} else {
					# Don't send any cookies but the ones we track (dump cookies provided by browser if any)
					delete $$headers{'cookie'};
					logevent ($user,$$headers{'x-forwarded-for'},"Host: $$self{'_hphf_proxy'}{'request'}{'_headers'}{'host'}\tCookie Note\tNo cookies to give.") if $logDetail;
					# print STDERR "No cookies to give.\n";
				}
			}
      }
   }
);

# The response header filter that follows handles receiving cookies from the 
# web server and storing them to be sent with later requests.

my $responseHeaderFilter = HTTP::Proxy::HeaderFilter::simple->new(
	sub { 
   	my ( $self, $headers, $message) = @_; 
   	
   	say ("MIME: $$headers{'content-type'}\n",4);
   	
      my ($hop, $i, $j, $cookiename, $value);

		my $uri = scalar($$message{'_request'}{'_uri'});
		my ($protocol, $hoststring, $url, $params) = $uri =~ m/([^:]*):\/\/([^\/]*)(\/[^?]*)?(\?.*)?/;

      if ($HoTTProxyConfig{'ProxyCookiesEnabled'}) {
	      foreach $i (sort(keys(%{$headers}))) { 
	        	if (lc($i) eq 'set-cookie') {
	        		my @headerList = ();
	        		if (ref($$headers{$i})) {
	        			@headerList = @{$$headers{$i}};
	        		} else {
	        			@headerList = ($$headers{$i});
	        		}
	     	   	foreach $j (@headerList) {
	     	   		say ("Caught Set-Cookie: $j\n",5);
	     	   		# print STDERR "$j\n";
	     	   		my %cookie = parseNetscapeCookie($j,$$message{'_request'}{'_headers'}{'host'},$url);
						logevent ($$self{'_hphf_proxy'}{'HoTTProxy-Request-User'},$$message{'_request'}{'_headers'}{'x-forwarded-for'},"Host: $$message{'_request'}{'_headers'}{'host'}\tSet-Cookie\t$j") if $logDetail;
	
	     	   		say ("Parsed Cookie: $cookie{'name'}, Domain: $cookie{'domain'}, Path: $cookie{'path'}, Expires: $cookie{'expires'}\n",5);

	     	   		# Check to see if we've already read in the bag of cookies for this domain restriction
	     	   		# If not, read it in
	     	   		if (not defined $$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}) {		# Cookie organization in memory is {'HoTTProxy-Cookie-Jar'}{username}{domain_restriction}{cookie_path/name}
	     	   			# print STDERR "Cookie bag wasn't in memory...\n";
	     	   			if (-e "$HoTTProxyConfig{'CookieDirectory'}/$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}\@$cookie{'domain'}.cookies") {		# format for cookie bag filename is username@domain_restriction.cookies
	     	   				# print STDERR "Cookie bag existed on disk...\n";
	     	   				my $fileHandle;
	     	   				if (open ($fileHandle, "$HoTTProxyConfig{'CookieDirectory'}/$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}\@$cookie{'domain'}.cookies")) {
	     	   					# print STDERR "Opened cookie bag on disk...\n";
	     	   					# print STDERR "Assigned fileHandle to array...\n";
	     	   					# print STDERR "Entering read loop...\n";
	     	   					while (<$fileHandle>) {
	     	   						chomp $_;
	     	   						if ($_) {
	     	   							my %tempCookie = parseStoredCookie($_);
	     	   							if ($tempCookie{'name'}) {
	     	   								my $cookieID = lc($tempCookie{'path'}) . lc($tempCookie{'name'});		# We're naming the cookie $path . $name so that cookies from the same path, domain, and name, will overwrite old ones
	     	   								$$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$tempCookie{'domain'}}{$cookieID} = {%tempCookie};
	     	   							}
	     	   						}
	     	   					}
	     	   					close $fileHandle;
	     	   				}
	     	   			}
	     	   		}
	     	   		# Now write the new cookie into memory
	     	   		my $cookieID = lc($cookie{'path'}) . lc($cookie{'name'});
	     	   		# print STDERR "Old: \n";
	     	   		# print STDERR "      Name: $$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID}{'name'}\n";
	     	   		# print STDERR "     Value: $$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID}{'value'}\n";
	     	   		# print STDERR "    Domain: $$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID}{'domain'}\n";
	     	   		# print STDERR "      Path: $$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID}{'path'}\n";
	     	   		# print STDERR "   Expires: $$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID}{'expires'}\n";
	     	   		# print STDERR "    Secure: $$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID}{'secure'}\n";
	     	   		# print STDERR "\n";
	
	     	   		if (not defined $$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{':=Updated=:'}) {			# Check to see if this cookie bag is already "hot" and needs to be updated
	     	   			# print STDERR ":=Updated=: was not defined\n";
	        	   		if (scalar(keys(%{$$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID}})) < 1) {	# Check to see if there was anything in the original cookie and if not this is new without any further checking needed
	     	   				say ("\nMade cookie bag $cookie{'domain'} hot.",5);
	     	   				$$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{':=Updated=:'} = 1;	# Note that an update to this cookie bag has occurred so we know to save it later
	     	   			} else {
	     	   				# print STDERR "The original key was defined.\n";
		        	   		foreach $i (keys(%{$$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID}})) {
		        	   			# print STDERR "NV $i\n";
		        	   			if ($$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID}{$i} ne $cookie{$i}) {
		        	   				say ("\nMade cookie bag $cookie{'domain'} hot.\n\n",5);
		        	   				$$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{':=Updated=:'} = 1;	# Note that an update to this cookie bag has occurred so we know to save it later
		        	   			} else {
		        	   				# Do nothing
		        	   				# print STDERR "$$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID}{$i} ne $cookie{$i}\n";
		        	   			}
		        	   		}
		        	   	}
	        	   	}
	
	     	   		$$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID} = {%cookie};
	     	   		# print STDERR "New: " . keys(%{$$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID}}) . "\n\n";
	     	   		# print STDERR "New: \n";
	     	   		# print STDERR "      Name: $$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID}{'name'}\n";
	     	   		# print STDERR "     Value: $$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID}{'value'}\n";
	     	   		# print STDERR "    Domain: $$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID}{'domain'}\n";
	     	   		# print STDERR "      Path: $$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID}{'path'}\n";
	     	   		# print STDERR "   Expires: $$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID}{'expires'}\n";
	     	   		# print STDERR "    Secure: $$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$cookie{'domain'}}{$cookieID}{'secure'}\n";
	     	   		# print STDERR "\n";
	     	   		
	        	   }
	        	   delete $$headers{$i};	# Don't send these cookies to the browser
	        	}
	      }
				
	      # Now we need to save all the cookie bags
	      foreach $i (keys(%{$$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}})) {
	      	# print STDERR "\$i is $i\n";
	      	if ($$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$i}{':=Updated=:'}) {
	      		say ("\nNeed to update cookie bag $$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}\@$i.cookies\n\n",5);
	         	if (open (my $fileHandle, ">$HoTTProxyConfig{'CookieDirectory'}/$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}\@$i.cookies")) {
	         		say ("   Overwriting cookie bag $$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}\@$i.cookies... ",5);
	         		foreach $j (keys(%{$$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$i}})) {
	         			if ($j ne ':=Updated=:') {		# :=Updated=: is a marker entry so we act on all but that key
	         				print $fileHandle assembleCookie($$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$i}{$j}) . "\n";
		         			# say ("   Saved cookie $j\n",5);
	         			}
	         		}
	         		close $fileHandle;
	         		say ("Done.\n\n",5);
	         	} else {
	         		print STDERR sortableDate() . " Failed to open cookie bag $HoTTProxyConfig{'CookieDirectory'}/$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}\@$i.cookies for output. Error: $!.\n";
	         	}
	         	# Need to delete these cookie bags as we finish since they persist forever (until the proxy server ends)
	         	delete $$self{'_hphf_proxy'}{'HoTTProxy-Cookie-Jar'}{$$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}}{$i};
	         	# print STDERR "Removed cookie bag $i from memory.\n";
	         } else {
	         	say ("   Cookie bag $$self{'_hphf_proxy'}{'HoTTProxy-Request-User'}\@$i.cookies is unchanged.\n\n",5);
	         }
	      
	     	# print STDERR "$message\n";
	   	# foreach my $xxx (sort(keys(%{$message}))) {
	   	# 	print STDERR "$xxx -> $$message{$xxx}\n";
	   	# }
	   	
	   	# print '-' x 79 . "\n\n";
	
	
	      
	      }
	   }
   }
);

# Needed to make sure the bogus proxy:homepage requests run our RHF that handles 
# authentication and local domain redirects
$proxy->push_filter( scheme => 'proxy', request => $requestHeaderFilter );

$proxy->push_filter( request => $requestHeaderFilter );

$proxy->push_filter( mime => '*/*', response => $responseHeaderFilter );

AddInFilters($proxy);

say (sortableDate() . " HoTTProxy $HoTTProxyVersion Started.\n\n",1);

$proxy->start;

sub proxyCookieSupport {
	# If you remove proxy cookie support from the program, remove this too and then
	# the 'ProxyCookiesEnabled' config setting will automatically go to zero
	return 1;
}

sub parseNetscapeCookie {
   my $cookie = shift;	# e.g. RCG01lastactivity=1119168277; domain=.homegunsmith.com; path=/; expires=Mon, 19-Jun-2006 08:04:37 GMT
   							# Note: should *not* have the Set-Cookie: prefix
   							
   my $host = shift;		# e.g. www.homegunsmith.com
   my $path = shift;		# e.g. /

   # if (not $cookie =~ m/^Set-Cookie:/i) {
   # 	return;	# Not a Netscape Set-Cookie header
   # }


   # // ***************  Rule validation for host names  *************** \\
   
   # Host name rules:
   
   # 1. Name must start with a period
   # 2. If name less than 2 periods then .local is appended

   # Rule #1
   if (not $host =~ m/^\./) {
   	$host = '.' . $host;	# make it so
   }
   
   # Rule #2
   if (periodCount($host) < 2) {
   	$host .= '.local';		# See RFC2109 to get an idea of what I'm doing here
   }

   # \\ ***************  Rule validation for host names  *************** //


   if ((not $cookie) or (not $host) or (not $path)) {
   	return;	# We can't properly figure out the cookie without all these things
   }
   
   # $cookie =~ s/^Set-Cookie://i;	# Remove the Set-Cookie: designator
   
   my @nvpair = split(/;/,$cookie);	# Split up the name/value pairs.  The first pair must be the cookie name and value
   
   my %cookie = ();
   my ($name, $value, $i, $lcname);
   
   # ($name, $value) = split(/=/,$nvpair[0]);
   ($name, $value) = $nvpair[0] =~ m/^([^=]*)=(.*)$/;				# Unlike the previous "split", this handles if there are equal signs ('=') embedded in the string
   $name =~ s/\s//g;	# remove any whitespace from the name
   $lcname = lc($name);		# we are going to use the lowercase name of the cookie for storage purposes
   
   $cookie{'name'} = $name;	# Later specs specify the cookie name is to be case insensitive, so when replacing old with new, we'll be case insensitive, but we'll store the case "as sent" so we can return it "as sent" in case the application cares.
   $cookie{'value'} = $value;
   
   for ($i=1; $i <=$#nvpair; $i++) {
   	# my($name,$value) = split(/=/,$nvpair[$i]);
   	my($name, $value) = $nvpair[$i] =~ m/^([^=]*)=(.*)$/;		# Unlike the previous "split", this handles if there are equal signs ('=') embedded in the string
   	$name =~ s/\s//g;	# remove any whitespace from the name
   	$name = lc($name);	# All of these parameters should have their names in lower case
   	if (($name eq 'domain') or ($name eq 'path') or ($name eq 'expires') or ($name eq 'secure')) {	# Only want the parameters we understand
   		if ($name eq 'secure') {		# The 'secure' nv pair isn't really a pair, only the name will appear, so we force a 1
   			$value = 1;
   		}
   		$cookie{$name} = $value;
   	}
   }

	if (not $cookie{'expires'}) {
		# Give the "session cookie" an expiration date because we don't have control of a "session"
		$cookie{'expiresEpoch'} = time + ($HoTTProxyConfig{'SessionCookieLife'} * 60);
		$cookie{'expires'} = time2str($cookie{'expiresEpoch'});
	} else {
		$cookie{'expiresEpoch'} = str2time($cookie{'expires'});
	}

   ## BBlakley: Note, probably want to throw some sort of log if any rule fails other than defaulting rules
   
   
   # // ***************  Rule validation for the path restriction  *************** \\
   
   # Path restriction rules:
   
   # 1. Path defaults to '/' if it is blank
   # 2. Path must start with a slash
   # 3. Path must end with a slash
   # 3. Path must head match the called path
   
   # Rule #1
   if (not $cookie{'path'}) {
   	$cookie{'path'} = '/';
   }
   
   # Rule #2
   if (not $cookie{'path'} =~ m/^\//) {		# if the first character of the path restriction isn't a slash
   	$cookie{'path'} = '/' . $cookie{'path'};	# make it so
   }

   # Rule #3
   if (not $cookie{'path'} =~ m/\/$/) {		# if the last character of the path restriction isn't a slash
   	$cookie{'path'} .= '/';	# make it so
   }

	# Rule #4
	if (not $path =~ m/^$cookie{'path'}/) {
		$cookie{'path'} = '/';
	}

   # \\ ***************  Rule validation for the path restriction  *************** //



   # // ***************  Rule validation for the domain restriction  *************** \\
   
   # Domain restriction rules:
   
   # 1. Domian *is* the host if it is blank
   # 2. Domain must start with a period
   # 3. Domain must have two periods
   # 4. Domain must tail match the host

	# Rule #1
   if (not $cookie{'domain'}) {
   	$cookie{'domain'} = $host;		# default domain if none specified
   }

	# Rule #2   
   if (not $cookie{'domain'} =~ m/^\./) {		# if the first character of the domain restriction isn't a period
   	$cookie{'domain'} = '.' . $cookie{'domain'};	# make it so
   }

	# Rule #3
	if (periodCount($cookie{'domain'}) < 2) {
		$cookie{'domain'} = $host;
	}
   
	# Rule #4
	if (not $host =~ m/$cookie{'domain'}$/) {
		$cookie{'domain'} = $host;
	}

   # \\ ***************  Rule validation for the domain restriction  *************** //

   
   return %cookie;
}

sub assembleCookie {
	# Legal hash items: name, value, domain, path, expires, secure
	my $cookieObject = shift;
	my ($i, $cookieText);

	$cookieText = "$$cookieObject{'name'}=$$cookieObject{'value'}; domain=$$cookieObject{'domain'}; path=$$cookieObject{'path'}";

	foreach $i ('expires', 'secure') {
   	if ($$cookieObject{$i}) {
   		$cookieText .= "; $i=$$cookieObject{$i}";
   	}
   }

	return $cookieText;	
}

sub parseStoredCookie {

	# returns hash of: $hash{'name'}
	# 						 $hash{'value'}
	# 						 $hash{'expires'}
	# 						 $hash{'expiresEpoch'}
	# 						 $hash{'domain'}
	# 						 $hash{'path'}
	# 						 $hash{'secure'}
   
   my $cookie = shift;	# e.g. RCG01lastactivity=1119168277; domain=.homegunsmith.com; path=/; expires=Mon, 19-Jun-2006 08:04:37 GMT
   							# Note: should *not* have the Set-Cookie: prefix
   							
   if (not $cookie) {
   	return;	# We can't properly figure out the cookie without all these things
   }

   my @nvpair = split(/;/,$cookie);	# Split up the name/value pairs.  The first pair must be the cookie name and value
   
   my %cookie = ();
   my ($name, $value, $i, $lcname);
   
   # ($name, $value) = split(/=/,$nvpair[0]);
   ($name, $value) = $nvpair[0] =~ m/^([^=]*)=(.*)$/;				# Unlike the previous "split", this handles if there are equal signs ('=') embedded in the string
   $name =~ s/\s//g;	# remove any whitespace from the name
   $lcname = lc($name);		# we are going to use the lowercase name of the cookie for storage purposes
   
   $cookie{'name'} = $name;	# Later specs specify the cookie name is to be case insensitive, so when replacing old with new, we'll be case insensitive, but we'll store the case "as sent" so we can return it "as sent" in case the application cares.
   $cookie{'value'} = $value;
   
   for ($i=1; $i <=$#nvpair; $i++) {
   	# my($name,$value) = split(/=/,$nvpair[$i]);
   	my($name, $value) = $nvpair[$i] =~ m/^([^=]*)=(.*)$/;				# Unlike the previous "split", this handles if there are equal signs ('=') embedded in the string
   	$name =~ s/\s//g;	# remove any whitespace from the name
   	$name = lc($name);	# All of these parameters should have their names in lower case
   	if (($name eq 'domain') or ($name eq 'path') or ($name eq 'expires') or ($name eq 'secure')) {	# Only want the parameters we understand
   		if ($name eq 'secure') {		# The 'secure' nv pair isn't really a pair, only the name will appear, so we force a 1
   			$value = 1;
   		}
   		$cookie{$name} = $value;
   	}
   }

	if ($cookie{'expires'}) {
		$cookie{'expiresEpoch'} = str2time($cookie{'expires'});
	}
   
   return %cookie;
}

sub sortableDate {
	# Used to return time as GMT.  Now just returning as local time without a time zone indicator
	# my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday) = gmtime(time);
	# return sprintf("%04d-%02d-%02d %02d:%02d:%02d GMT",$year+1900,$mon+1,$mday,$hour,$min,$sec);
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday) = localtime(time);
	return sprintf("%04d-%02d-%02d %02d:%02d:%02d",$year+1900,$mon+1,$mday,$hour,$min,$sec);
}

sub logevent {
	my $user = shift;
	my $ip = shift;
	my $message = shift;
	print OUT sortableDate() . "\t$user\t$ip\t$message\n";
	autoflush OUT;
}

sub periodCount {			# How many periods ('.') in a string?
	my $string = shift;
	$string =~ s/[^.]//g;
	return length($string);
}

sub loadConfig {
	my $configFile = shift;
	my $filehandle;
	if (not open ($filehandle,"$path/$configFile")) {
		die "Cannot open config file $configFile.\nError: $!\nEnding.\n";
	}
	while (<$filehandle>) {
		chomp $_;
		if ($_ =~ m/^#/) {		# if the line starts with a # character
			# do nothing, this is a comment line
		} elsif ($_) {
			my ($name,$value) = $_ =~ m/^([^=]*) = (.*)$/;	# e.g. DefaultHomePage = http://wap.google.com
			if ($name =~ s/^-//) {
				$hiddenConfig{$name} = 1;	# elements with names that start with a minus sign will have the minus sign stripped, but will be hidden from the console display - great for things that never change and nobody cares about
			}
			if ($name =~ m/s$/) {	# If the element name ends in 's' force it to be a "multi" (e.g. TrustedDomains or ValidLocalHosts) - done so if a multi has only one value, it is still created as a multi
				my @values = split(/;/,$value);	# Note: don't put spaces around your ; delimiters in a value list or they'll end up in the values
				foreach my $i (@values) {
					$HoTTProxyConfig{$name}{$i} = 1;		# Just make them true so they exist and can easily be tested for
				}
			} else {
				$HoTTProxyConfig{$name} = $value;
			}
		}
	}
	close $filehandle;
	if (not $HoTTProxyConfig{'SessionCookieLife'}) {
		$HoTTProxyConfig{'SessionCookieLife'} = 1440;	# Default to 1 day if not specified
	}
	if ($HoTTProxyConfig{'CookiePassthru'}) {
		$HoTTProxyConfig{'CookiePassthru'} = 'Feature not yet written';
	}
	if ($HoTTProxyConfig{'SetCookiePassthru'}) {
		$HoTTProxyConfig{'SetCookiePassthru'} = 'Feature not yet written';
	}
}

sub processParameters {
	my $i;
	for ($i=0; $i <= $#ARGV; $i++) {
		my $p = $ARGV[$i];
		next if not $p =~ m/^\//;	# Only process parameters that start with frontslash, others are parameters to parameters

		if ($p eq '/?') {
			syntax();
		} elsif ($p eq '/license') {
			license();
		} elsif ($p eq '/logdetail') {
			$HoTTProxyConfig{'LogDetail'} = 1;
		} elsif ($p eq '/config') {
			$HoTTProxyConfig{'ConfigFile'} = $ARGV[$i+1];
			die "No configuration file name followed /config parameter. Ending.\n\n" if not $HoTTProxyConfig{'ConfigFile'};
		} elsif ($p eq '/q') {
			$HoTTProxyConfig{'ConsoleVerbosity'} = 0;
		} elsif ($p eq '/v') {
			$HoTTProxyConfig{'ConsoleVerbosity'} = $ARGV[$i+1];
			die "No verbosity level followed /v parameter. Ending.\n\n" if not $HoTTProxyConfig{'ConsoleVerbosity'};
		}
	}
}

sub filefree {
	my $filename = shift;
	
	# Here we want to find out if a given file is "free" or if some other process has it open.
	# The only reliable way I've found to do that on Win32 is to try to rename the file. If 
	# the rename fails, the file is in use, otherwise it is "free".
	
	# rename returns true on success or false on failure - perfect!
	my $filefree = rename($filename,"$filename.}filefree{");
	
	if ($filefree) {
		# Need to rename it back!!!
		rename("$filename.}filefree{",$filename);
	}
	
	return $filefree;
}

sub syntax {
	say("

Syntax:

HoTTProxy [/?] [/logdetail] [/config myconfigfile.ext] [/license] ([/q] | /v x)

  /?          Shows this help
  /logdetail  More detail in HoTTProxyLog.txt (LogDetail*)
  /config     Tells HoTTProxy to use a config file other than HoTTProxy.conf
  /license    Displays the license for the software
  /q          Set ConsoleVerbosity to zero*
  /v          Set ConsoleVerbosity any value other than zero*

* = Overrides the stated value in the .conf file.

HoTTProxy is a HTTP Proxy server. Operational parameters are in the 
configuration file. I see no reason why you couldn't run multiple instances 
of HoTTProxy concurrently on a single machine, each 'listening' to a different 
port (if you wanted to support multiple ports).

Note that HoTTProxy requires users to authenticate. To create user accounts 
create a text file named username.password in the same directory as the 
HoTTProxy executable, replacing the word 'username' with the desired username. 
Inside that file, make the first line be the desired password, and the second 
line be the user's preferred homepage -- to which they will be redirected any 
time they try to access a 'local' domain like http://homepage, etc.  The second 
line (the homepage) is optional and they will be directed to the default 
homepage specified in the configuration file if they have no personal setting.

See the HoTTProxyDocs.txt file for more details.

",0);

	exit;
}

sub say {
	my $message = shift;
	my $messageLevel = shift;
	
	if ($messageLevel <= $HoTTProxyConfig{'ConsoleVerbosity'}) {
		print $message;
	}
}

sub license {
	if (not open(IN,"$path/license.txt")) {
		say ("
Unable to open License.txt for input.

Your copy of this program is either incomplete or this file is missing for 
another reason.  Obtain an original, complete copy of the software from 
http://www.HoTTProxy.org.

",0);
	} else {
		my @license = <IN>;
		close IN;
		say (join('',@license),0);
	}
	exit;
}
